import psycopg2.extensions
from django.db import OperationalError, connection
from django.db.transaction import DEFAULT_DB_ALIAS, Atomic
from psycopg2.errorcodes import SERIALIZATION_FAILURE


def serializable_atomic(using=None, savepoint=True):
    return atomic(using, savepoint, psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE)


def atomic(using=None, savepoint=True, isolation_level=psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT):
    if callable(using):
        # Bare decorator: @atomic
        # although the first argument is called `using`, it's actually the function being decorated.
        return AtomicWrapper(DEFAULT_DB_ALIAS, savepoint, isolation_level)(using)
    else:
        # Parametric decorator: @atomic(...) or context manager: with atomic(...): ...
        return AtomicWrapper(using, savepoint, isolation_level)


class AtomicWrapper(Atomic):
    _level_names = {
        psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT: 'AUTOCOMMIT',
        psycopg2.extensions.ISOLATION_LEVEL_READ_UNCOMMITTED: 'READ UNCOMMITTED',
        psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED: 'READ COMMITTED',
        psycopg2.extensions.ISOLATION_LEVEL_REPEATABLE_READ: 'REPEATABLE READ',
        psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE: 'SERIALIZABLE',
    }

    def __init__(self, using, savepoint, isolation_level):
        super(AtomicWrapper, self).__init__(using, savepoint)
        self.isolation_level = isolation_level
        self.isolation_level_name = AtomicWrapper._level_names[isolation_level]

    def __enter__(self):
        super(AtomicWrapper, self).__enter__()
        cursor = connection.cursor()
        cursor.execute('SET TRANSACTION ISOLATION LEVEL {}'.format(self.isolation_level_name))

    def __exit__(self, exc_type, exc_val, exc_tb):
        result = super(AtomicWrapper, self).__exit__(exc_type, exc_val, exc_tb)
        if result:
            return result
        if exc_type == OperationalError:
            # Get exc_val.__cause__.pgcode
            pgcode = getattr(getattr(exc_val, '__cause__', None), 'pgcode', '')
            if pgcode == SERIALIZATION_FAILURE:
                raise ConcurrencyError()
        return result


class ConcurrencyError(Exception):
    pass



from django.forms import BaseForm
from django.utils.translation import get_language
from rest_framework.serializers import Serializer

try:
    from urlparse import parse_qs, urlparse
except ImportError:  # For Python 3
    from urllib.parse import parse_qs, urlparse


def django_files_to_requests_files(files):
    return {k: (f.name, f.read(), f.content_type) for k, f in files.items()} if files is not None else None


def to_error_list(serializer_or_form):
    """
    Converts from django error format to a list of error strings. The fields must have `label` attribute.

    :param serializer_or_form: django serializer or form with failed validation,
     i.e. serializer_or_form.is_valid() is False.
    :return: A list of error strings
    """
    assert isinstance(serializer_or_form, Serializer) or isinstance(serializer_or_form, BaseForm), (
        'Input parameter must be a rest_framework serializer or a django form.'
    )

    return [
        '{field_name}: {error_message}'.format(
            field_name=serializer_or_form.fields[field_key].label,
            error_message=field_errors[0],
        )
        for field_key, field_errors in serializer_or_form.errors.items()
    ]


def append_lang_param(url, param_name):
    query = urlparse(url).query
    if param_name not in parse_qs(query):
        if url[-1] not in ['?', '&']:
            url += '&' if query else '?'
        url += '{}={}'.format(param_name, get_language())
    return url



"""
Django REST framework handles the following exceptions, and automatically generates error responses:
  - Subclasses of APIException raised inside REST framework,
  - Django's Http404 exception,
  - and Django's PermissionDenied exception.

More info: http://www.django-rest-framework.org/api-guide/exceptions/

We capture them here, and reformat responses to our desired style.
"""
import logging
import json

from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, status
from rest_framework.response import Response
from rest_framework.views import exception_handler

from django_on_chain import SETTINGS
from django_on_chain.exceptions import ServiceError

logger = logging.getLogger(__name__)


def reformat_error_handler(e, context):
    if isinstance(e, ServiceError):
        content = {
            'type': e.type,
            'message': e.message,
        }
        if e.errors is not None:
            content['errors'] = e.errors
        return Response(content, status=e.suggested_http_status or status.HTTP_500_INTERNAL_SERVER_ERROR)

    response = exception_handler(e, context)  # Call default exception handler to get the standard error response

    if response is None:
        traceback = e.__traceback__ if hasattr(e, '__traceback__') else None
        logger.error('Uncaught exception', exc_info=(type(e), e, traceback), extra={
            'request': __serialize_request(context['request']),
        })

    reformat_errors = getattr(context.get('view', object()), 'reformat_errors', SETTINGS['REFORMAT_ALL_ERRORS'])

    if not reformat_errors:
        return response

    if response is None:
        return Response({
            'type': 'server_error',
            'message': _('Unknown error; please try again later'),
        }, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    if response.status_code == status.HTTP_400_BAD_REQUEST:
        message = _('Invalid request')
        if isinstance(e, exceptions.ValidationError):
            message += ': {}'.format(e.detail)
        response.data = {
            'type': 'bad_request',
            'message': message,
        }
    elif response.status_code == status.HTTP_404_NOT_FOUND:
        response.data = {
            'type': 'not_found',
            'message': _('Resource was not found.'),
        }
    elif response.status_code == status.HTTP_401_UNAUTHORIZED:
        response.data = {
            'type': 'unauthorized',
            'message': _('Request is not authorized'),
        }
    elif response.status_code == status.HTTP_403_FORBIDDEN:
        message = _('You do not have permission to perform this action.')
        if isinstance(e, exceptions.APIException):
            message = e.detail
        response.data = {
            'type': 'access_forbidden',
            'message': message,
        }
    elif response.status_code == status.HTTP_429_TOO_MANY_REQUESTS:
        response.data = {
            'type': 'too_many_requests',
            'message': _("Too many requests"),
        }
    elif isinstance(e, exceptions.APIException):
        response.data = {
            'type': e.default_code,
            'message': force_text(e.default_detail),
        }

    return response


def __serialize_request(request):
    try:
        result = {
            'scheme': request.scheme,
            'url': request.path,
            'method': request.method,

            'content_type': request.content_type,
            'content_params': request.content_params,

            'GET': {key: values for (key, values) in request.GET.lists()},
            'POST': '',
            'COOKIES': request.COOKIES,
            'META': {key: value for key, value in request.META.items() if key.isupper()},

            'FILES': {
                key: [
                    {
                        'name': f.name,
                        'size': f.size,
                        'content_type': f.content_type,
                    }
                    for f in files
                ]
                for (key, files) in request.FILES.lists()
            },
        }

    except Exception:
        logger.exception('django-on-chain exception while serializing request object')
        result = request

    return result



from django.conf import settings

SETTINGS = {
    'DEFAULT_REQUEST_TIMEOUT': 10,
    'REFORMAT_ALL_ERRORS': True,
}

SETTINGS.update(getattr(settings, 'DJANGO_ON_CHAIN', {}))

if SETTINGS.get('DEFAULT_TIMEOUT'):
    # For backward compatibility until all clients are upgraded
    SETTINGS['DEFAULT_REQUEST_TIMEOUT'] = SETTINGS['DEFAULT_TIMEOUT']
